package speiger.src.coreengine.rendering.gui.font;

import java.util.Map;
import java.util.function.Consumer;

import speiger.src.coreengine.assets.AssetLocation;
import speiger.src.coreengine.rendering.gui.font.glyth.Glyth;
import speiger.src.coreengine.rendering.gui.font.glyth.GlythData;
import speiger.src.coreengine.rendering.gui.font.glyth.MissingGlyth;
import speiger.src.coreengine.rendering.gui.font.glyth.UnbakedGlyth.GlythBaker;
import speiger.src.coreengine.rendering.tesselation.buffer.IVertexBuilder;

public class Font {
	private static final AssetLocation DEFAULT = AssetLocation.of("default");
	
	Map<AssetLocation, FontGroup> fonts;
	GlythBaker baker;
	float size;
	float oversample;
	GlythCache[] styledCache = new GlythCache[] {new GlythCache(this, 0), new GlythCache(this, 1), new GlythCache(this, 2), new GlythCache(this, 3)};
	MissingGlyth missingGlyth;
	GlythData missingData;
	
	protected Font(Map<AssetLocation, FontGroup> fonts, GlythBaker baker, float size, float oversample, Consumer<Runnable> clearing) {
		this.fonts = fonts;
		this.baker = baker;
		this.size = size;
		this.oversample = oversample;
		this.missingGlyth = new MissingGlyth(size, oversample * 2F);
		missingData = new GlythData(missingGlyth);
		clearing.accept(this::reset);
	}
	
	private void reset() {
		for(int i = 0;i<4;i++) {
			styledCache[i].reset();
		}
		missingGlyth.cleanCache();
		missingData.cleanCache();
	}
	
	public GlythData data(AssetLocation font, int codepoint, int style) {
		return styledCache[style & 0x3].data(font, codepoint);
	}
	
	public Glyth glyth(AssetLocation font, int codepoint, int style) {
		return styledCache[style & 0x3].glyth(font, codepoint);
	}
	
	protected FontGroup font(AssetLocation font) {
		return fonts.get(font);
	}
	
	public void drawText(String text, float x, float y, int color, TexturedBuffer buffer) {
		int previousCodepoint = -1;
		for(int i = 0,m=text.length();i<m;) {
			int codepoint = text.codePointAt(i);
			GlythData data = data(DEFAULT, codepoint, 0);
			if(previousCodepoint != -1) {
				x -= data.kerning(previousCodepoint);
			}
			Glyth glyth = glyth(DEFAULT, codepoint, 0);
			if(glyth.isValid()) {
				float minX = glyth.left() + x;
				float minY = glyth.top() + y;
				float maxX = glyth.right() + x;
				float maxY = glyth.bottom() + y;
				
				IVertexBuilder builder = buffer.builderForTexture(glyth.texture());
				builder.pos(minX, minY, 0F).tex(glyth.minU(), glyth.minV()).rgba(-1).endVertex();
				builder.pos(maxX, minY, 0F).tex(glyth.maxU(), glyth.minV()).rgba(-1).endVertex();
				builder.pos(maxX, maxY, 0F).tex(glyth.maxU(), glyth.maxV()).rgba(-1).endVertex();
				builder.pos(maxX, maxY, 0F).tex(glyth.maxU(), glyth.maxV()).rgba(-1).endVertex();
				builder.pos(minX, maxY, 0F).tex(glyth.minU(), glyth.maxV()).rgba(-1).endVertex();
				builder.pos(minX, minY, 0F).tex(glyth.minU(), glyth.minV()).rgba(-1).endVertex();
			}
			x += data.advance();
			i += Character.charCount(codepoint);
			previousCodepoint = codepoint;
		}
		buffer.end();
	}
	
	public static interface TexturedBuffer {
		public IVertexBuilder builderForTexture(int textureId);
		public void end();
	}
}
